==> net.bluemind.sds.store.scalityring-5.4.2760-sources.jar: net/bluemind/sds/store/scalityring/ScalityRingStore.java <==
/* BEGIN LICENSE
 * Copyright © Blue Mind SAS, 2012-2019
 *
 * This file is part of BlueMind. BlueMind is a messaging and collaborative
 * solution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU Affero General Public License as
 * published by the Free Software Foundation (version 3 of the License).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See LICENSE.txt
 * END LICENSE
 */
package net.bluemind.sds.store.scalityring;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.LongAdder;

import org.asynchttpclient.AsyncCompletionHandler;
import org.asynchttpclient.AsyncHttpClient;
import org.asynchttpclient.HttpResponseBodyPart;
import org.asynchttpclient.Response;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.io.CountingInputStream;
import com.netflix.spectator.api.Clock;
import com.netflix.spectator.api.Counter;
import com.netflix.spectator.api.DistributionSummary;
import com.netflix.spectator.api.Registry;
import com.netflix.spectator.api.Timer;

import net.bluemind.common.io.Buffered;
import net.bluemind.metrics.registry.IdFactory;
import net.bluemind.sds.dto.DeleteRequest;
import net.bluemind.sds.dto.ExistRequest;
import net.bluemind.sds.dto.ExistResponse;
import net.bluemind.sds.dto.GetRequest;
import net.bluemind.sds.dto.MgetRequest;
import net.bluemind.sds.dto.MgetRequest.Transfer;
import net.bluemind.sds.dto.PutRequest;
import net.bluemind.sds.dto.SdsError;
import net.bluemind.sds.dto.SdsResponse;
import net.bluemind.sds.dto.TierMoveRequest;
import net.bluemind.sds.dto.TierMoveResponse;
import net.bluemind.sds.store.ISdsBackingStore;
import net.bluemind.sds.store.scalityring.zstd.ZstdStreams;

public class ScalityRingStore implements ISdsBackingStore {
	private static final Logger logger = LoggerFactory.getLogger(ScalityRingStore.class);

	@SuppressWarnings("serial")
	public static final class ClientAlreadyClosedException extends RuntimeException {
		public ClientAlreadyClosedException(String message) {
			super(message);
		}
	}

	private final int parallelism;
	private final AsyncHttpClient client;
	private final String endpoint;

	private final IdFactory idFactory;
	private final Timer getLatencyTimer;
	private final Timer mgetLatencyTimer;
	private final Timer existLatencyTimer;
	private final Timer putLatencyTimer;
	private final Timer deleteLatencyTimer;
	private final Clock clock;
	private final Counter getSizeCounter;
	private final Counter existRequestCounter;
	private final Counter existFailureRequestCounter;
	private final Counter getRequestCounter;
	private final Counter getFailureRequestCounter;
	private final Counter putRequestCounter;
	private final Counter putFailureRequestCounter;
	private final Counter mgetRequestCounter;
	private final Counter deleteRequestCounter;
	private final Counter deleteFailureRequestCounter;
	private final Counter putSizeCounter;
	private DistributionSummary compressionRatio;

	public ScalityRingStore(ScalityConfiguration configuration, AsyncHttpClient client, int parallelism,
			Registry registry, IdFactory idfactory) {
		this.parallelism = parallelism;
		this.idFactory = idfactory;
		this.client = client;
		endpoint = configuration.endpoint;

		clock = registry.clock();
		getSizeCounter = registry.counter(idFactory.name("transfer").withTag("direction", "download"));
		getRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "get").withTag("status", "success"));
		getFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "get").withTag("status", "error"));
		putRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "put").withTag("status", "success"));
		putFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "put").withTag("status", "error"));
		existRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "exist").withTag("status", "success"));
		existFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "exist").withTag("status", "error"));
		mgetRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "mget").withTag("status", "success"));
		deleteRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "delete").withTag("status", "success"));
		deleteFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "delete").withTag("status", "error"));
		putSizeCounter = registry.counter(idFactory.name("transfer").withTag("direction", "upload"));
		existLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "exist"));
		mgetLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "mget"));
		getLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "get"));
		putLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "put"));
		compressionRatio = registry.distributionSummary(idFactory.name("compressionRatio"));
		deleteLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "delete"));
	}

	@Override
	public CompletableFuture<ExistResponse> exists(ExistRequest req) {
		final long start = clock.monotonicTime();
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}
		return client.prepareHead(endpoint + "/" + req.guid) //
				// Makes no attempt to return object metadata,
				// and the “X-Scal-Usermd” header is not included in the response.
				// This may avoid some disk reads when system metadata is in memory.
				.addHeader("X-Scal-Get-Usermd", "No") //
				.execute() //
				.toCompletableFuture() //
				.thenApply(httpresp -> {
					existLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
					boolean known = httpresp.getStatusCode() == 200;
					Counter requestCounter = known ? existRequestCounter : existFailureRequestCounter;
					requestCounter.increment();
					return ExistResponse.from(known);
				}).exceptionally(t -> {
					logger.error("exists request failed", t);
					existFailureRequestCounter.increment();
					return ExistResponse.from(false);
				});
	}

	@Override
	public CompletableFuture<SdsResponse> upload(PutRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		File uploadFile = new File(req.filename);
		long uploadLength = uploadFile.length();

		return exists(ExistRequest.of(req.guid)) //
				.thenApply(er -> er.exists).thenCompose(existresponse -> {
					SdsResponse sr = new SdsResponse();
					sr.withTags(Map.of("guid", req.guid, "skip", "true"));
					if (Boolean.TRUE.equals(existresponse)) {
						return CompletableFuture.completedFuture(sr);
					}
					CountingInputStream stream = ZstdStreams.compress(uploadFile);
					return client.preparePut(endpoint + "/" + req.guid) //
							.setBody(stream) //
							.execute() //
							.toCompletableFuture() //
							.thenApply(httpresp -> {
								boolean success = httpresp.getStatusCode() == 200;
								putLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
								putSizeCounter.increment(uploadLength);
								long compressionPercent = uploadLength != 0 ? stream.getCount() * 100 / uploadLength
										: 100;
								compressionRatio.record(Math.max(0, 100 - compressionPercent));
								Counter requestCounter = success ? putRequestCounter : putFailureRequestCounter;
								requestCounter.increment();
								if (!success) {
									String statusMessage = httpresp.getStatusText();
									sr.error = new SdsError(statusMessage != null ? statusMessage : "no response text",
											httpresp.getStatusCode() != 404);
								}
								return sr;
							});
				}).exceptionally(t -> {
					logger.error("put request failed", t);
					SdsResponse sr = SdsResponse.error(t.getMessage(), req.guid, true);
					sr.withTags(Map.of("guid", req.guid));
					putFailureRequestCounter.increment();
					return sr;
				});
	}

	@Override
	public CompletableFuture<SdsResponse> download(GetRequest req) {
		Path downloadPath = Paths.get(req.filename);
		return download(req, downloadPath, ZstdStreams.decompress(req.guid, downloadPath));
	}

	@Override
	public CompletableFuture<SdsResponse> downloadRaw(GetRequest req) {
		Path downloadPath = Paths.get(req.filename);
		try {
			return download(req, downloadPath, Buffered.output(Files.newOutputStream(downloadPath)));
		} catch (IOException e) {
			logger.error("download local io failed", e);
			SdsResponse sr = SdsResponse.error(e.getMessage(), req.guid, true);
			sr.withTags(Map.of("guid", req.guid));
			deleteFailureRequestCounter.increment();
			return CompletableFuture.completedFuture(sr);
		}
	}

	public CompletableFuture<SdsResponse> download(GetRequest req, Path downloadPath, OutputStream outChannel) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		CompletableFuture<SdsResponse> response = new CompletableFuture<>();
		AtomicLong downloaded = new AtomicLong(0);

		client.prepareGet(endpoint + "/" + req.guid) //
				.execute(new AsyncCompletionHandler<Void>() {
					@Override
					public State onBodyPartReceived(HttpResponseBodyPart bodyPart) throws IOException {
						byte[] data = bodyPart.getBodyPartBytes();
						long len = data.length;
						if (len > 0) {
							downloaded.addAndGet(len);
							outChannel.write(data);
						}
						return State.CONTINUE;
					}

					@Override
					public void onThrowable(Throwable t) {
						try {
							outChannel.close();
						} catch (IOException e) {
						}
						logger.error("download request for {} failed", req.guid, t);
						getFailureRequestCounter.increment();
						response.completeExceptionally(t);
					}

					@Override
					public Void onCompleted(Response httpresp) throws Exception {
						boolean success = httpresp.getStatusCode() == 200;
						outChannel.close();
						if (!success) {
							getFailureRequestCounter.increment();
							response.completeExceptionally(new RuntimeException("failed to download " + req.guid + ": "
									+ httpresp.getStatusCode() + " " + httpresp.getStatusText()));
							// Don't store the '404 error' page on the local filesystem
							Files.deleteIfExists(downloadPath);
							return null;
						}
						getLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
						getRequestCounter.increment();
						getSizeCounter.increment(downloaded.get());
						response.complete(new SdsResponse().withSize(downloaded.get()));
						return null;
					}
				});
		return response;
	}

	private CompletableFuture<SdsResponse> delete0(String guid) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		return client.prepareDelete(endpoint + "/" + guid) //
				// Makes no attempt to return object metadata,
				// and the “X-Scal-Usermd” header is not included in the response.
				// This may avoid some disk reads when system metadata is in memory.
				.execute() //
				.toCompletableFuture() //
				.thenApply(httpresp -> {
					boolean success = httpresp.getStatusCode() == 200;
					deleteLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
					Counter requestCounter = success ? deleteRequestCounter : deleteFailureRequestCounter;
					requestCounter.increment();
					return SdsResponse.UNTAGGED_OK;
				}).exceptionally(t -> {
					logger.error("delete request failed", t);
					SdsResponse sr = SdsResponse.error(t.getMessage(), guid, false);
					sr.withTags(Map.of("guid", guid));
					deleteFailureRequestCounter.increment();
					return sr;
				}).thenApply(sdsr -> {
					logger.info("sdsresp {}", sdsr);
					return sdsr;
				});
	}

	@Override
	public CompletableFuture<SdsResponse> delete(DeleteRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		List<CompletableFuture<SdsResponse>> futures = req.guids.stream().parallel().map(this::delete0).toList();
		return CompletableFuture.allOf(futures.toArray(CompletableFuture[]::new)).thenApply(v -> {
			List<SdsResponse> sdserrors = futures.stream().map(CompletableFuture::join)
					.filter(sdsr -> !sdsr.succeeded()).toList();
			if (sdserrors.size() >= futures.size()) {
				logger.error("All delete failed: {}", sdserrors);
				return SdsResponse.error("All downloads failed. First one: " + sdserrors.getFirst().error, req.guids,
						true);
			}
			if (!sdserrors.isEmpty()) {
				logger.warn("some deletion failed: {}", sdserrors);
			}
			return SdsResponse.UNTAGGED_OK;
		});
	}

	@Override
	public CompletableFuture<SdsResponse> downloads(MgetRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();

		int len = req.transfers.size();
		final LongAdder totalSize = new LongAdder();
		int parallelStreams = Math.max(128, parallelism);
		CompletableFuture<?>[] roots = new CompletableFuture[parallelStreams];
		for (int i = 0; i < parallelStreams; i++) {
			roots[i] = CompletableFuture.completedFuture(null);
		}
		Iterator<Transfer> it = req.transfers.iterator();
		for (int i = 0; i < len; i++) {
			int slot = i % parallelStreams;
			Transfer t = it.next();

			roots[slot] = roots[slot]
					.thenCompose(v -> download(GetRequest.of(req.mailbox, t.guid, t.filename)).thenApply(sdsresp -> {
						totalSize.add(sdsresp.size());
						return sdsresp;
					}));
		}

		return CompletableFuture.allOf(roots).thenApply(v -> {
			mgetLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
			getSizeCounter.increment(totalSize.longValue());
			mgetRequestCounter.increment();
			String sizeKb = Long.toString(totalSize.longValue() / 1024);
			return new SdsResponse().withTags(Map.of("batch", Integer.toString(len), "sizeKB", sizeKb));
		}).exceptionally(ex -> {
			logger.error(ex.getMessage() + " for " + req, ex);
			return SdsResponse.error(ex.getMessage(), req.transfers.stream().map(t -> t.guid).toList(), true);
		});
	}

	@Override
	public CompletableFuture<TierMoveResponse> tierMove(TierMoveRequest tierMoveRequest) {
		return CompletableFuture.completedFuture(new TierMoveResponse(
				tierMoveRequest.moves.stream().map(tm -> tm.messageBodyGuid).toList(), Collections.emptyList()));
	}

	@Override
	public void close() {
		if (client != null && !client.isClosed()) {
			try {
				client.close();
			} catch (IOException e) {
				logger.error("Unable to close scality ring store", e);
			}
		}
	}

}

==> net.bluemind.sds.store.scalityring-5.5.3184-sources.jar: net/bluemind/sds/store/scalityring/ScalityRingStore.java <==
/* BEGIN LICENSE
 * Copyright © Blue Mind SAS, 2012-2019
 *
 * This file is part of BlueMind. BlueMind is a messaging and collaborative
 * solution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU Affero General Public License as
 * published by the Free Software Foundation (version 3 of the License).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See LICENSE.txt
 * END LICENSE
 */
package net.bluemind.sds.store.scalityring;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.LongAdder;

import org.asynchttpclient.AsyncCompletionHandler;
import org.asynchttpclient.AsyncHttpClient;
import org.asynchttpclient.HttpResponseBodyPart;
import org.asynchttpclient.Response;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.io.CountingInputStream;
import com.netflix.spectator.api.Clock;
import com.netflix.spectator.api.Counter;
import com.netflix.spectator.api.DistributionSummary;
import com.netflix.spectator.api.Registry;
import com.netflix.spectator.api.Timer;

import net.bluemind.metrics.registry.IdFactory;
import net.bluemind.sds.dto.DeleteRequest;
import net.bluemind.sds.dto.ExistRequest;
import net.bluemind.sds.dto.ExistResponse;
import net.bluemind.sds.dto.GetRequest;
import net.bluemind.sds.dto.MgetRequest;
import net.bluemind.sds.dto.MgetRequest.Transfer;
import net.bluemind.sds.dto.PutRequest;
import net.bluemind.sds.dto.SdsError;
import net.bluemind.sds.dto.SdsResponse;
import net.bluemind.sds.dto.TierMoveRequest;
import net.bluemind.sds.dto.TierMoveResponse;
import net.bluemind.sds.store.ISdsBackingStore;
import net.bluemind.sds.store.PathHandler;
import net.bluemind.sds.store.scalityring.zstd.ZstdStreams;

public class ScalityRingStore implements ISdsBackingStore {
	private static final Logger logger = LoggerFactory.getLogger(ScalityRingStore.class);

	@SuppressWarnings("serial")
	public static final class ClientAlreadyClosedException extends RuntimeException {
		public ClientAlreadyClosedException(String message) {
			super(message);
		}
	}

	private final int parallelism;
	private final AsyncHttpClient client;
	private final String endpoint;

	private final IdFactory idFactory;
	private final Timer getLatencyTimer;
	private final Timer mgetLatencyTimer;
	private final Timer existLatencyTimer;
	private final Timer putLatencyTimer;
	private final Timer deleteLatencyTimer;
	private final Clock clock;
	private final Counter getSizeCounter;
	private final Counter existRequestCounter;
	private final Counter existFailureRequestCounter;
	private final Counter getRequestCounter;
	private final Counter getFailureRequestCounter;
	private final Counter putRequestCounter;
	private final Counter putFailureRequestCounter;
	private final Counter mgetRequestCounter;
	private final Counter deleteRequestCounter;
	private final Counter deleteFailureRequestCounter;
	private final Counter putSizeCounter;
	private DistributionSummary compressionRatio;

	public ScalityRingStore(ScalityConfiguration configuration, AsyncHttpClient client, int parallelism,
			Registry registry, IdFactory idfactory) {
		this.parallelism = parallelism;
		this.idFactory = idfactory;
		this.client = client;
		endpoint = configuration.endpoint;

		clock = registry.clock();
		getSizeCounter = registry.counter(idFactory.name("transfer").withTag("direction", "download"));
		getRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "get").withTag("status", "success"));
		getFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "get").withTag("status", "error"));
		putRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "put").withTag("status", "success"));
		putFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "put").withTag("status", "error"));
		existRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "exist").withTag("status", "success"));
		existFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "exist").withTag("status", "error"));
		mgetRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "mget").withTag("status", "success"));
		deleteRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "delete").withTag("status", "success"));
		deleteFailureRequestCounter = registry
				.counter(idFactory.name("request").withTag("method", "delete").withTag("status", "error"));
		putSizeCounter = registry.counter(idFactory.name("transfer").withTag("direction", "upload"));
		existLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "exist"));
		mgetLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "mget"));
		getLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "get"));
		putLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "put"));
		compressionRatio = registry.distributionSummary(idFactory.name("compressionRatio"));
		deleteLatencyTimer = registry.timer(idFactory.name("latency").withTag("method", "delete"));
	}

	@Override
	public CompletableFuture<ExistResponse> exists(ExistRequest req) {
		final long start = clock.monotonicTime();
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}
		return client.prepareHead(endpoint + "/" + req.guid) //
				// Makes no attempt to return object metadata,
				// and the “X-Scal-Usermd” header is not included in the response.
				// This may avoid some disk reads when system metadata is in memory.
				.addHeader("X-Scal-Get-Usermd", "No") //
				.execute() //
				.toCompletableFuture() //
				.thenApply(httpresp -> {
					existLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
					boolean known = httpresp.getStatusCode() == 200;
					Counter requestCounter = known ? existRequestCounter : existFailureRequestCounter;
					requestCounter.increment();
					return ExistResponse.from(known);
				}).exceptionally(t -> {
					logger.error("exists request failed", t);
					existFailureRequestCounter.increment();
					return ExistResponse.from(false);
				});
	}

	@Override
	public CompletableFuture<SdsResponse> upload(PutRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		PathHandler uploadFile = PathHandler.forPathOrUri(req.filename);
		long uploadLength = uploadFile.size();

		return exists(ExistRequest.of(req.guid)) //
				.thenApply(er -> er.exists).thenCompose(existresponse -> {
					SdsResponse sr = new SdsResponse();
					sr.withTags(Map.of("guid", req.guid, "skip", "true"));
					if (Boolean.TRUE.equals(existresponse)) {
						return CompletableFuture.completedFuture(sr);
					}
					CountingInputStream stream = ZstdStreams.compress(uploadFile);
					return client.preparePut(endpoint + "/" + req.guid) //
							.setBody(stream) //
							.execute() //
							.toCompletableFuture() //
							.thenApply(httpresp -> {
								boolean success = httpresp.getStatusCode() == 200;
								putLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
								putSizeCounter.increment(uploadLength);
								long compressionPercent = uploadLength != 0 ? stream.getCount() * 100 / uploadLength
										: 100;
								compressionRatio.record(Math.max(0, 100 - compressionPercent));
								Counter requestCounter = success ? putRequestCounter : putFailureRequestCounter;
								requestCounter.increment();
								if (!success) {
									String statusMessage = httpresp.getStatusText();
									sr.error = new SdsError(statusMessage != null ? statusMessage : "no response text",
											httpresp.getStatusCode() != 404);
								}
								return sr;
							});
				}).exceptionally(t -> {
					logger.error("put request failed", t);
					SdsResponse sr = SdsResponse.error(t.getMessage(), req.guid, true);
					sr.withTags(Map.of("guid", req.guid));
					putFailureRequestCounter.increment();
					return sr;
				});
	}

	@Override
	public CompletableFuture<SdsResponse> download(GetRequest req) {
		PathHandler downloadPath = PathHandler.forPathOrUri(req.filename);
		return download(req, downloadPath, ZstdStreams.decompress(req.guid, downloadPath));
	}

	@Override
	public CompletableFuture<SdsResponse> downloadRaw(GetRequest req) {
		PathHandler downloadPath = PathHandler.forPathOrUri(req.filename);
		try {
			return download(req, downloadPath, downloadPath.openForWriting());
		} catch (IOException e) {
			logger.error("download local io failed", e);
			SdsResponse sr = SdsResponse.error(e.getMessage(), req.guid, true);
			sr.withTags(Map.of("guid", req.guid));
			deleteFailureRequestCounter.increment();
			return CompletableFuture.completedFuture(sr);
		}
	}

	public CompletableFuture<SdsResponse> download(GetRequest req, PathHandler downloadPath, OutputStream outChannel) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		CompletableFuture<SdsResponse> response = new CompletableFuture<>();
		AtomicLong downloaded = new AtomicLong(0);
		AtomicBoolean ignoreParts = new AtomicBoolean(false);

		client.prepareGet(endpoint + "/" + req.guid) //
				.execute(new AsyncCompletionHandler<Void>() {
					@Override
					public State onBodyPartReceived(HttpResponseBodyPart bodyPart) throws IOException {
						if (!ignoreParts.get()) {
							byte[] data = bodyPart.getBodyPartBytes();
							long len = data.length;
							if (len > 0) {
								downloaded.addAndGet(len);
								outChannel.write(data);
							}
						}
						return State.CONTINUE;
					}

					@Override
					public void onThrowable(Throwable t) {
						try {
							outChannel.close();
						} catch (IOException e) {
						}
						logger.error("download request for {} failed", req.guid, t);
						getFailureRequestCounter.increment();
						response.completeExceptionally(t);
					}

					@Override
					public Void onCompleted(Response httpresp) throws Exception {
						boolean success = httpresp.getStatusCode() == 200;
						outChannel.close();
						if (!success) {
							ignoreParts.set(true);
							getFailureRequestCounter.increment();
							response.completeExceptionally(new RuntimeException("failed to download " + req.guid + ": "
									+ httpresp.getStatusCode() + " " + httpresp.getStatusText()));
							// Don't store the '404 error' page on the local filesystem
							downloadPath.cleanup();
							return null;
						}
						getLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
						getRequestCounter.increment();
						getSizeCounter.increment(downloaded.get());
						response.complete(new SdsResponse().withSize(downloaded.get()));
						return null;
					}
				});
		return response;
	}

	private CompletableFuture<SdsResponse> delete0(String guid) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();
		return client.prepareDelete(endpoint + "/" + guid) //
				// Makes no attempt to return object metadata,
				// and the “X-Scal-Usermd” header is not included in the response.
				// This may avoid some disk reads when system metadata is in memory.
				.execute() //
				.toCompletableFuture() //
				.thenApply(httpresp -> {
					boolean success = httpresp.getStatusCode() == 200;
					deleteLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
					Counter requestCounter = success ? deleteRequestCounter : deleteFailureRequestCounter;
					requestCounter.increment();
					return SdsResponse.UNTAGGED_OK;
				}).exceptionally(t -> {
					logger.error("delete request failed", t);
					SdsResponse sr = SdsResponse.error(t.getMessage(), guid, false);
					sr.withTags(Map.of("guid", guid));
					deleteFailureRequestCounter.increment();
					return sr;
				}).thenApply(sdsr -> {
					logger.info("sdsresp {}", sdsr);
					return sdsr;
				});
	}

	@Override
	public CompletableFuture<SdsResponse> delete(DeleteRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		List<CompletableFuture<SdsResponse>> futures = req.guids.stream().parallel().map(this::delete0).toList();
		return CompletableFuture.allOf(futures.toArray(CompletableFuture[]::new)).thenApply(v -> {
			List<SdsResponse> sdserrors = futures.stream().map(CompletableFuture::join)
					.filter(sdsr -> !sdsr.succeeded()).toList();
			if (sdserrors.size() >= futures.size()) {
				logger.error("All delete failed: {}", sdserrors);
				return SdsResponse.error("All downloads failed. First one: " + sdserrors.getFirst().error, req.guids,
						true);
			}
			if (!sdserrors.isEmpty()) {
				logger.warn("some deletion failed: {}", sdserrors);
			}
			return SdsResponse.UNTAGGED_OK;
		});
	}

	@Override
	public CompletableFuture<SdsResponse> downloads(MgetRequest req) {
		if (client.isClosed()) {
			return CompletableFuture.failedFuture(new ClientAlreadyClosedException("NO: client is closed"));
		}

		final long start = clock.monotonicTime();

		int len = req.transfers.size();
		final LongAdder totalSize = new LongAdder();
		int parallelStreams = Math.max(128, parallelism);
		CompletableFuture<?>[] roots = new CompletableFuture[parallelStreams];
		for (int i = 0; i < parallelStreams; i++) {
			roots[i] = CompletableFuture.completedFuture(null);
		}
		Iterator<Transfer> it = req.transfers.iterator();
		for (int i = 0; i < len; i++) {
			int slot = i % parallelStreams;
			Transfer t = it.next();

			roots[slot] = roots[slot]
					.thenCompose(v -> download(GetRequest.of(req.mailbox, t.guid, t.filename)).thenApply(sdsresp -> {
						totalSize.add(sdsresp.size());
						return sdsresp;
					}));
		}

		return CompletableFuture.allOf(roots).thenApply(v -> {
			mgetLatencyTimer.record(clock.monotonicTime() - start, TimeUnit.NANOSECONDS);
			getSizeCounter.increment(totalSize.longValue());
			mgetRequestCounter.increment();
			String sizeKb = Long.toString(totalSize.longValue() / 1024);
			return new SdsResponse().withTags(Map.of("batch", Integer.toString(len), "sizeKB", sizeKb));
		}).exceptionally(ex -> {
			logger.error(ex.getMessage() + " for " + req, ex);
			return SdsResponse.error(ex.getMessage(), req.transfers.stream().map(t -> t.guid).toList(), true);
		});
	}

	@Override
	public CompletableFuture<TierMoveResponse> tierMove(TierMoveRequest tierMoveRequest) {
		return CompletableFuture.completedFuture(new TierMoveResponse(
				tierMoveRequest.moves.stream().map(tm -> tm.messageBodyGuid).toList(), Collections.emptyList()));
	}

	@Override
	public void close() {
		if (client != null && !client.isClosed()) {
			try {
				client.close();
			} catch (IOException e) {
				logger.error("Unable to close scality ring store", e);
			}
		}
	}

}
